/* mulmod-k8-fpu.h -- (C) Geoffrey Reynolds, July 2006.

   Inline modular multiplication routines for x86-64/FPU.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#if defined(NEED_mulmod64) && !defined(HAVE_mulmod64)
#define HAVE_mulmod64

#define FPU_MODE_BITS "(0x0F00)"  /* 64-bit precision, round to zero. */

#define MOD64_MAX_BITS 62

/* Saved copy of FPU mode bits. */
extern uint16_t mod64_rnd;

/* a*b (mod p), where a,b < p < 2^62.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains 1.0/p computed with above settings.
*/
static inline uint64_t mulmod64(uint64_t a, uint64_t b, uint64_t p)
{
  register uint64_t ret, r1;
  uint64_t tmp;

  asm ("fildll  %3"                 "\n\t"
       "fildll  %4"                 "\n\t"
       "fmulp   %%st(0), %%st(1)"   "\n\t"
       "fmul    %%st(1), %%st(0)"   "\n\t"
       "fistpll %2"                 "\n\t"
       "imul    %1, %0"             "\n\t"
       "mov     %2, %1"             "\n\t"
       "imul    %5, %1"             "\n\t"
       "sub     %1, %0"             "\n\t"
       "mov     %0, %1"             "\n\t"
       "sub     %5, %1"             "\n\t"
       "jl      0f"                 "\n\t"
       "mov     %1, %0"             "\n"
       "0:"
       : "=r" (ret), "=r" (r1), "=m" (tmp)
       : "m" (a), "m" (b), "rm" (p), "0" (a), "1" (b), "m" (mod64_rnd)
       : "cc" );

  return ret;
}

/* Sets FPU into double extended precision and round-to-zero mode.
   Pushes 1.0/p onto the FPU stack, computed with above settings.
   Saves the old mode bits in mod64_rnd.
*/
static inline void mod64_init(uint64_t p)
{
  uint16_t tmp;

  asm volatile ("fildll  %2"               "\n\t"
                "fld1"                     "\n\t"
                "fnstcw  %0"               "\n\t"
                "mov     %0, %%ax"         "\n\t"
                "and     $"FPU_MODE_BITS", %%ax"    "\n\t"
                "orw     $"FPU_MODE_BITS", %0"      "\n\t"
                "mov     %%ax, %1"         "\n\t"  /* Save old bits */
                "fldcw   %0"               "\n\t"
                "fdivp   %%st(0),%%st(1)"

                : "=m" (tmp), "+m" (mod64_rnd)
                : "m" (p)
                : "%rax", "cc" );
}

/* Pops 1.0/p off the FPU stack and restores the previous FPU mode
   from mod64_rnd.
*/
static inline void mod64_fini(void)
{
  uint16_t tmp;

  asm volatile ("fstp    %%st(0)"          "\n\t"
                "fnstcw  %0"               "\n\t"
                "mov     %0, %%ax"         "\n\t"
                "and     $~"FPU_MODE_BITS", %%ax"    "\n\t"
                "or      %1, %%ax"         "\n\t"  /* Restore old bits */
                "mov     %%ax, %0"         "\n\t"
                "fldcw   %0"

                : "=m" (tmp), "+m" (mod64_rnd)
                : 
                : "%rax", "cc" );
}

/* a*b (mod p), where a,b < p < 2^62.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define PRE2_MULMOD64(a,b,p) ({ \
  register uint64_t ret, r1; \
  uint64_t tmp; \
  asm ("fildll  %3"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "imul    %1, %0"             "\n\t" \
       "mov     %2, %1"             "\n\t" \
       "imul    %5, %1"             "\n\t" \
       "sub     %1, %0"             "\n\t" \
       "mov     %0, %1"             "\n\t" \
       "sub     %5, %1"             "\n\t" \
       "jl      0f"                 "\n\t" \
       "mov     %1, %0"             "\n" \
       "0:" \
       : "=r" (ret), "=r" (r1), "=m" (tmp) \
       : "m" (a), "0" (b), "rm" (p), "1" (a), "m" (mod64_rnd) \
       : "cc" ); \
  ret; })

/* Pushes b/p onto the FPU stack. Assumes %st(0) contains 1.0/p.
 */
#define PRE2_MULMOD64_INIT(b) \
  asm volatile ("fildll  %1\n\t" \
                "fmul    %%st(1),%%st(0)" \
                : "+m" (mod64_rnd) \
                : "m" ((uint64_t)b) \
                : "cc" )

/* Pops b/p off the FPU stack.
 */
#define PRE2_MULMOD64_FINI() \
  asm volatile ("fstp    %%st(0)" \
                : "+m" (mod64_rnd) \
                : \
                : "cc" )

#endif /* defined(NEED_mulmod64) && !defined(HAVE_mulmod64) */


#if defined(NEED_sqrmod64) && !defined(HAVE_sqrmod64)
#define HAVE_sqrmod64
extern uint16_t mod64_rnd;

/* b^2 (mod p), where b < p < 2^62.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains 1.0/p computed with above settings.
*/
static inline uint64_t sqrmod64(uint64_t a, uint64_t p)
{
  register uint64_t ret, r1;
  uint64_t tmp;

  asm ("fildll  %3"                 "\n\t"
       "fmul    %%st(0), %%st(0)"   "\n\t"
       "fmul    %%st(1), %%st(0)"   "\n\t"
       "fistpll %2"                 "\n\t"
       "imul    %0, %0"             "\n\t"
       "mov     %2, %1"             "\n\t"
       "imul    %4, %1"             "\n\t"
       "sub     %1, %0"             "\n\t"
       "mov     %0, %1"             "\n\t"
       "sub     %4, %1"             "\n\t"
       "jl      0f"                 "\n\t"
       "mov     %1, %0"             "\n"
       "0:"
       : "=r" (ret), "=&r" (r1), "=m" (tmp)
       : "m" (a), "rm" (p), "0" (a), "m" (mod64_rnd)
       : "cc" );

  return ret;
}
#endif /* defined(NEED_sqrmod64) && !defined(HAVE_sqrmod64) */


#if defined(NEED_powmod64) && !defined(HAVE_powmod64)
#define HAVE_powmod64
/*
  b^n (mod p)
*/
uint64_t powmod64_k8_fpu(uint64_t b, uint64_t n, uint64_t p)
     attribute ((pure));
#define powmod64(b,n,p) powmod64_k8_fpu(b,n,p)
#endif


#if defined(NEED_vector) && !defined(HAVE_vector)
#define HAVE_vector

/* Given uint64_t X[2], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i+2] <-- X[i] * b (mod p), for 0 <= i < 2.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC2_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  uint64_t T[2] __attribute__ ((aligned(16))); \
 \
  asm ("fildll  %10"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %11"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3"                 "\n\t" \
 \
       "mov     %10, %4"   "\n\t" \
       "mov     %11, %5"   "\n\t" \
       "imul    %8, %4"    "\n\t" \
       "imul    %8, %5"    "\n\t" \
       "mov     %2, %6"    "\n\t" \
       "mov     %3, %7"    "\n\t" \
       "imul    %9, %6"    "\n\t" \
       "imul    %9, %7"    "\n\t" \
       "sub     %6, %4"    "\n\t" \
       "sub     %7, %5"    "\n\t" \
       "mov     %4, %0"    "\n\t" \
       "mov     %5, %1"    "\n\t" \
       "sub     %9, %4"    "\n\t" \
       "jl      0f"        "\n\t" \
       "mov     %4, %0"    "\n"   \
       "0:"                "\t"   \
       "sub     %9, %5"    "\n\t" \
       "jl      1f"        "\n\t" \
       "mov     %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[2]), "=m" ((X)[3]), "=m" (T[0]), "=m" (T[1]), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "m" ((X)[0]), "m" ((X)[1]), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[2], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 2.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC2_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  uint64_t T[2] __attribute__ ((aligned(16))); \
 \
  asm ("fildll  %0"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %1"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3"                 "\n\t" \
 \
       "mov     %0, %4"    "\n\t" \
       "mov     %1, %5"    "\n\t" \
       "imul    %8, %4"    "\n\t" \
       "imul    %8, %5"    "\n\t" \
       "mov     %2, %6"    "\n\t" \
       "mov     %3, %7"    "\n\t" \
       "imul    %9, %6"    "\n\t" \
       "imul    %9, %7"    "\n\t" \
       "sub     %6, %4"    "\n\t" \
       "sub     %7, %5"    "\n\t" \
       "mov     %4, %0"    "\n\t" \
       "mov     %5, %1"    "\n\t" \
       "sub     %9, %4"    "\n\t" \
       "jl      0f"        "\n\t" \
       "mov     %4, %0"    "\n"   \
       "0:"                "\t"   \
       "sub     %9, %5"    "\n\t" \
       "jl      1f"        "\n\t" \
       "mov     %5, %1"    "\n"   \
       "1:"                         \
       : "+m" ((X)[0]), "+m" ((X)[1]), "=m" (T[0]), "=m" (T[1]), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[4], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i+4] <-- X[i] * b (mod p), for 0 <= i < 4.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC4_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, t0, t1; \
  uint64_t T[2] __attribute__ ((aligned(16))); \
 \
  asm ("fildll  %10"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %11"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3"                 "\n\t" \
 \
       "mov     %10, %4"   "\n\t" \
       "mov     %11, %5"   "\n\t" \
       "imul    %8, %4"    "\n\t" \
       "imul    %8, %5"    "\n\t" \
       "mov     %2, %6"    "\n\t" \
       "mov     %3, %7"    "\n\t" \
       "imul    %9, %6"    "\n\t" \
       "imul    %9, %7"    "\n\t" \
       "sub     %6, %4"    "\n\t" \
       "sub     %7, %5"    "\n\t" \
       "mov     %4, %0"    "\n\t" \
       "mov     %5, %1"    "\n\t" \
       "sub     %9, %4"    "\n\t" \
       "jl      0f"        "\n\t" \
       "mov     %4, %0"    "\n"   \
       "0:"                "\t"   \
       "sub     %9, %5"    "\n\t" \
       "jl      1f"        "\n\t" \
       "mov     %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[4]), "=m" ((X)[5]), "=m" (T[0]), "=m" (T[1]), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "m" ((X)[0]), "m" ((X)[1]), "m" (mod64_rnd) \
       : "cc" ); \
 \
  asm ("fildll  %10"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %11"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3"                 "\n\t" \
 \
       "mov     %10, %4"   "\n\t" \
       "mov     %11, %5"   "\n\t" \
       "imul    %8, %4"    "\n\t" \
       "imul    %8, %5"    "\n\t" \
       "mov     %2, %6"    "\n\t" \
       "mov     %3, %7"    "\n\t" \
       "imul    %9, %6"    "\n\t" \
       "imul    %9, %7"    "\n\t" \
       "sub     %6, %4"    "\n\t" \
       "sub     %7, %5"    "\n\t" \
       "mov     %4, %0"    "\n\t" \
       "mov     %5, %1"    "\n\t" \
       "sub     %9, %4"    "\n\t" \
       "jl      0f"        "\n\t" \
       "mov     %4, %0"    "\n"   \
       "0:"                "\t"   \
       "sub     %9, %5"    "\n\t" \
       "jl      1f"        "\n\t" \
       "mov     %5, %1"    "\n"   \
       "1:"                         \
       : "=m" ((X)[6]), "=m" ((X)[7]), "=m" (T[0]), "=m" (T[1]), \
         "=&r" (r0), "=&r" (r1), "=&r" (t0), "=&r" (t1) \
       : "rm" (b), "rm" (p), "m" ((X)[2]), "m" ((X)[3]), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[4], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 4.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC4_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  uint64_t T[4] __attribute__ ((aligned(32))); \
 \
  asm ("fildll  %0"                  "\n\t" \
       "fmul    %%st(1), %%st(0)"    "\n\t" \
       "fistpll %4"                  "\n\t" \
       "fildll  %1"                  "\n\t" \
       "fmul    %%st(1), %%st(0)"    "\n\t" \
       "fistpll %5"                  "\n\t" \
 \
       "fildll  %2"                  "\n\t" \
       "fmul    %%st(1), %%st(0)"    "\n\t" \
       "fistpll %6"                  "\n\t" \
       "fildll  %3"                  "\n\t" \
       "fmul    %%st(1), %%st(0)"    "\n\t" \
       "fistpll %7"                  "\n\t" \
 \
       "mov     %0, %8"     "\n\t" \
       "mov     %1, %9"     "\n\t" \
       "imul    %16, %8"    "\n\t" \
       "imul    %16, %9"    "\n\t" \
       "mov     %4, %12"    "\n\t" \
       "mov     %5, %13"    "\n\t" \
       "imul    %17, %12"   "\n\t" \
       "imul    %17, %13"   "\n\t" \
       "sub     %12, %8"    "\n\t" \
       "sub     %13, %9"    "\n\t" \
 \
       "mov     %2, %10"    "\n\t" \
       "mov     %3, %11"    "\n\t" \
       "imul    %16, %10"   "\n\t" \
       "imul    %16, %11"   "\n\t" \
       "mov     %6, %14"    "\n\t" \
       "mov     %7, %15"    "\n\t" \
       "imul    %17, %14"   "\n\t" \
       "imul    %17, %15"   "\n\t" \
       "sub     %14, %10"   "\n\t" \
       "sub     %15, %11"   "\n\t" \
 \
       "mov     %8, %0"     "\n\t" \
       "mov     %9, %1"     "\n\t" \
       "sub     %17, %8"    "\n\t" \
       "jl      0f"         "\n\t" \
       "mov     %8, %0"     "\n"   \
       "0:"                 "\t"   \
       "sub     %17, %9"    "\n\t" \
       "jl      1f"         "\n\t" \
       "mov     %9, %1"     "\n"   \
       "1:"                 "\t"   \
 \
       "mov     %10, %2"    "\n\t" \
       "mov     %11, %3"    "\n\t" \
       "sub     %17, %10"   "\n\t" \
       "jl      0f"         "\n\t" \
       "mov     %10, %2"    "\n"   \
       "0:"                 "\t"   \
       "sub     %17, %11"   "\n\t" \
       "jl      1f"         "\n\t" \
       "mov     %11, %3"    "\n"   \
       "1:" \
       : "+m" ((X)[0]), "+m" ((X)[1]), "+m" ((X)[2]), "+m" ((X)[3]), \
         "=m" (T[0]), "=m" (T[1]), "=m" (T[2]), "=m" (T[3]), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), "m" (mod64_rnd) \
       : "cc" ); \
}

/* Given uint64_t X[8], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i+8] <-- X[i] * b (mod p), for 0 <= i < 8.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC8_MULMOD64_NEXT(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  uint64_t T[8] __attribute__ ((aligned(16))); \
 \
  asm ("fildll  %4"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %0"                 "\n\t" \
       "fildll  %5"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %1"                 "\n\t" \
       "fildll  %6"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %7"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3" \
 \
       : "=m" (T[0]), "=m" (T[1]), "=m" (T[2]), "=m" (T[3]) \
       : "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), "m" ((X)[3]), \
         "m" (mod64_rnd) ); \
 \
  asm ("mov       %14, %4"   "\n\t" \
       "mov       %15, %5"   "\n\t" \
       "mov       %16, %6"   "\n\t" \
       "mov       %17, %7"   "\n\t" \
       "mov       %18, %8"   "\n\t" \
       "mov       %19, %9"   "\n\t" \
       "mov       %20, %10"  "\n\t" \
       "mov       %21, %11"  "\n\t" \
       "imul      %12, %4"   "\n\t" \
       "imul      %12, %5"   "\n\t" \
       "imul      %12, %6"   "\n\t" \
       "imul      %12, %7"   "\n\t" \
       "imul      %13, %8"   "\n\t" \
       "imul      %13, %9"   "\n\t" \
       "imul      %13, %10"  "\n\t" \
       "imul      %13, %11"  "\n\t" \
       "sub       %8, %4"    "\n\t" \
       "sub       %9, %5"    "\n\t" \
       "sub       %10, %6"   "\n\t" \
       "sub       %11, %7"   "\n\t" \
 \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "mov       %6, %2"    "\n\t" \
       "mov       %7, %3"    "\n\t" \
       "sub       %13, %4"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %13, %5"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %13, %6"   "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %6, %2"    "\n"   \
       "2:"                  "\t"   \
       "sub       %13, %7"   "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %7, %3"    "\n"   \
       "3:"                         \
       : "=m" ((X)[8]), "=m" ((X)[9]), "=m" ((X)[10]), "=m" ((X)[11]), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), \
         "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), "m" ((X)[3]), \
         "m" (T[0]), "m" (T[1]), "m" (T[2]), "m" (T[3]) \
       : "cc" ); \
 \
  asm ("fildll  %4"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %0"                 "\n\t" \
       "fildll  %5"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %1"                 "\n\t" \
       "fildll  %6"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %7"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3" \
 \
       : "=m" (T[4]), "=m" (T[5]), "=m" (T[6]), "=m" (T[7]) \
       : "m" ((X)[4]), "m" ((X)[5]), "m" ((X)[6]), "m" ((X)[7]), \
         "m" (mod64_rnd) ); \
 \
  asm ("mov       %14, %4"   "\n\t" \
       "mov       %15, %5"   "\n\t" \
       "mov       %16, %6"   "\n\t" \
       "mov       %17, %7"   "\n\t" \
       "mov       %18, %8"   "\n\t" \
       "mov       %19, %9"   "\n\t" \
       "mov       %20, %10"  "\n\t" \
       "mov       %21, %11"  "\n\t" \
       "imul      %12, %4"   "\n\t" \
       "imul      %12, %5"   "\n\t" \
       "imul      %12, %6"   "\n\t" \
       "imul      %12, %7"   "\n\t" \
       "imul      %13, %8"   "\n\t" \
       "imul      %13, %9"   "\n\t" \
       "imul      %13, %10"  "\n\t" \
       "imul      %13, %11"  "\n\t" \
       "sub       %8, %4"    "\n\t" \
       "sub       %9, %5"    "\n\t" \
       "sub       %10, %6"   "\n\t" \
       "sub       %11, %7"   "\n\t" \
 \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "mov       %6, %2"    "\n\t" \
       "mov       %7, %3"    "\n\t" \
       "sub       %13, %4"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %13, %5"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %13, %6"   "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %6, %2"    "\n"   \
       "2:"                  "\t"   \
       "sub       %13, %7"   "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %7, %3"    "\n"   \
       "3:"                         \
       : "=m" ((X)[12]), "=m" ((X)[13]), "=m" ((X)[14]), "=m" ((X)[15]), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), \
         "m" ((X)[4]), "m" ((X)[5]), "m" ((X)[6]), "m" ((X)[7]), \
         "m" (T[4]), "m" (T[5]), "m" (T[6]), "m" (T[7]) \
       : "cc" ); \
}

/* Given uint64_t X[8], b, p, where 0 <= X[i],b < p < 2^62:
   Assign X[i] <-- X[i] * b (mod p), for 0 <= i < 8.
   Assumes FPU is set to double extended precision and round-to-zero.
   Assumes %st(0) contains b/p computed with above settings.
*/
#define VEC8_MULMOD64(X,b,p) { \
  register uint64_t r0, r1, r2, r3, t0, t1, t2, t3; \
  uint64_t T[8] __attribute__ ((aligned(16))); \
 \
  asm ("fildll  %8"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %0"                 "\n\t" \
       "fildll  %9"                 "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %1"                 "\n\t" \
       "fildll  %10"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %2"                 "\n\t" \
       "fildll  %11"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %3"                 "\n\t" \
       "fildll  %12"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %4"                 "\n\t" \
       "fildll  %13"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %5"                 "\n\t" \
       "fildll  %14"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %6"                 "\n\t" \
       "fildll  %15"                "\n\t" \
       "fmul    %%st(1), %%st(0)"   "\n\t" \
       "fistpll %7" \
 \
       : "=m" (T[0]), "=m" (T[1]), "=m" (T[2]), "=m" (T[3]), \
         "=m" (T[4]), "=m" (T[5]), "=m" (T[6]), "=m" (T[7]) \
       : "m" ((X)[0]), "m" ((X)[1]), "m" ((X)[2]), "m" ((X)[3]), \
         "m" ((X)[4]), "m" ((X)[5]), "m" ((X)[6]), "m" ((X)[7]), \
         "m" (mod64_rnd) ); \
 \
  asm ("mov       %0, %4"    "\n\t" \
       "mov       %1, %5"    "\n\t" \
       "mov       %2, %6"    "\n\t" \
       "mov       %3, %7"    "\n\t" \
       "mov       %14, %8"   "\n\t" \
       "mov       %15, %9"   "\n\t" \
       "mov       %16, %10"  "\n\t" \
       "mov       %17, %11"  "\n\t" \
       "imul      %12, %4"   "\n\t" \
       "imul      %12, %5"   "\n\t" \
       "imul      %12, %6"   "\n\t" \
       "imul      %12, %7"   "\n\t" \
       "imul      %13, %8"   "\n\t" \
       "imul      %13, %9"   "\n\t" \
       "imul      %13, %10"  "\n\t" \
       "imul      %13, %11"  "\n\t" \
       "sub       %8, %4"    "\n\t" \
       "sub       %9, %5"    "\n\t" \
       "sub       %10, %6"   "\n\t" \
       "sub       %11, %7"   "\n\t" \
 \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "mov       %6, %2"    "\n\t" \
       "mov       %7, %3"    "\n\t" \
       "sub       %13, %4"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %13, %5"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %13, %6"   "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %6, %2"    "\n"   \
       "2:"                  "\t"   \
       "sub       %13, %7"   "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %7, %3"    "\n"   \
       "3:"                         \
       : "+m" ((X)[0]), "+m" ((X)[1]), "+m" ((X)[2]), "+m" ((X)[3]), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), \
         "m" (T[0]), "m" (T[1]), "m" (T[2]), "m" (T[3]) \
       : "cc" ); \
 \
  asm ("mov       %0, %4"    "\n\t" \
       "mov       %1, %5"    "\n\t" \
       "mov       %2, %6"    "\n\t" \
       "mov       %3, %7"    "\n\t" \
       "mov       %14, %8"   "\n\t" \
       "mov       %15, %9"   "\n\t" \
       "mov       %16, %10"  "\n\t" \
       "mov       %17, %11"  "\n\t" \
       "imul      %12, %4"   "\n\t" \
       "imul      %12, %5"   "\n\t" \
       "imul      %12, %6"   "\n\t" \
       "imul      %12, %7"   "\n\t" \
       "imul      %13, %8"   "\n\t" \
       "imul      %13, %9"   "\n\t" \
       "imul      %13, %10"  "\n\t" \
       "imul      %13, %11"  "\n\t" \
       "sub       %8, %4"    "\n\t" \
       "sub       %9, %5"    "\n\t" \
       "sub       %10, %6"   "\n\t" \
       "sub       %11, %7"   "\n\t" \
 \
       "mov       %4, %0"    "\n\t" \
       "mov       %5, %1"    "\n\t" \
       "mov       %6, %2"    "\n\t" \
       "mov       %7, %3"    "\n\t" \
       "sub       %13, %4"   "\n\t" \
       "jl        0f"        "\n\t" \
       "mov       %4, %0"    "\n"   \
       "0:"                  "\t"   \
       "sub       %13, %5"   "\n\t" \
       "jl        1f"        "\n\t" \
       "mov       %5, %1"    "\n"   \
       "1:"                  "\t"   \
       "sub       %13, %6"   "\n\t" \
       "jl        2f"        "\n\t" \
       "mov       %6, %2"    "\n"   \
       "2:"                  "\t"   \
       "sub       %13, %7"   "\n\t" \
       "jl        3f"        "\n\t" \
       "mov       %7, %3"    "\n"   \
       "3:"                         \
       : "+m" ((X)[4]), "+m" ((X)[5]), "+m" ((X)[6]), "+m" ((X)[7]), \
         "=&r" (r0), "=&r" (r1), "=&r" (r2), "=&r" (r3), \
         "=&r" (t0), "=&r" (t1), "=&r" (t2), "=&r" (t3) \
       : "rm" (b), "rm" (p), \
         "m" (T[4]), "m" (T[5]), "m" (T[6]), "m" (T[7]) \
       : "cc" ); \
}
#endif /* defined(NEED_vector) && !defined(HAVE_vector) */
