/* subseq.c -- (C) Geoffrey Reynolds, June 2006.

   Base b sequences in n=qm+r where n0 < n < n1 are represented as a number
   of base b^Q subsequences in m where n0/Q < m < n1/Q.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <stdlib.h>
#include "sr5sieve.h"
#include "bitmap.h"

subseq_t *SUBSEQ = NULL;
uint32_t subseq_count = 0;
uint32_t subseq_Q = 0;
uint32_t factor_count;        /* candidate n eliminated so far */
int duplicates_opt;

static uint32_t m_low;
static uint32_t m_high;

static uint_fast32_t subseq_clear_m(uint32_t subseq, uint32_t m)
{
  assert(subseq < subseq_count);

  if (m < m_low || m > m_high)
    return 0;

#if UPDATE_BITMAPS
  return clear_bit(SUBSEQ[subseq].M, m-m_low);
#else
  return test_bit(SUBSEQ[subseq].M, m-m_low);
#endif
}

/* Return 1 iff subsequence h of k*b^n+c has any terms with n%a==b.
 */ 
static int congruent_terms(uint32_t h, uint32_t a, uint32_t b)
{
  uint_fast32_t i, j;
  uint32_t g;

  assert(b < a);

  g = gcd32(a,subseq_Q);
  if (b % g == SUBSEQ[h].d % g)
  {
    j = m_high-m_low;
    for (i = first_bit(SUBSEQ[h].M); i <= j; i = next_bit(SUBSEQ[h].M,i+1))
      if (((i+m_low)*subseq_Q+SUBSEQ[h].d) % a == b)
        return 1;
  }

  return 0;
}
/* Build tables sc_lists[i][j] of pointers to lists of subsequences
   whose terms k*b^m+c satisfy m = j (mod r) for i = divisor_index[r].
*/
static void make_subseq_congruence_tables(void)
{
  uint32_t h, i, j, k, r, len;
  uint16_t subseq_list[POWER_RESIDUE_LCM];

  for (i = 0; i < seq_count; i++)
  {
    for (j = 1, r = 0; j <= POWER_RESIDUE_LCM; j++)
      if (POWER_RESIDUE_LCM % j == 0)
      {
        SEQ[i].sc_lists[r] = xmalloc(j*sizeof(uint16_t *));
        for (k = 0; k < j; k++)
        {
          uint16_t *tmp;

          for (len = 0, h = SEQ[i].first; h <= SEQ[i].last; h++)
            if (congruent_terms(h,j,k))
              subseq_list[len++] = h;
          if (len == 0)
            tmp = NULL;
          else
          {
            tmp = xmalloc((len+1)*sizeof(uint16_t));
            for (h = 0; h < len; h++)
              tmp[h] = subseq_list[h];
            tmp[h] = SUBSEQ_MAX;
          }
          SEQ[i].sc_lists[r][k] = tmp;
        }
        r++;
      }
    assert(r == POWER_RESIDUE_DIVISORS);
  }
}

void make_subseqs(void)
{
  uint32_t i, j, r, n, s, s_count;
  subseq_t *tmp;

#ifndef NDEBUG
  uint32_t t_count = 0, old_count = 0;
#endif

  subseq_Q = find_best_Q(&s_count);
  tmp = xmalloc(s_count*sizeof(subseq_t));
  m_low = n_min/subseq_Q;
  m_high = n_max/subseq_Q;

  for (i = 0, s = 0; i < seq_count; i++)
  {
    uint32_t nmin[LIMIT_BASE], subseq[LIMIT_BASE];
#if CHECK_FOR_GFN
    uint32_t g[LIMIT_BASE];
#endif

    SEQ[i].first = s;

    for (j = 0; j < subseq_Q; j++)
    {
#if CHECK_FOR_GFN
      g[j] = 0;
#endif
      nmin[j] = UINT32_MAX;
    }

    for (j = 0; j < SEQ[i].ncount; j++)
    {
      n = SEQ[i].N[j];
      r = n % subseq_Q;
      nmin[r] = MIN(nmin[r],n);
#if CHECK_FOR_GFN
      g[r] = gcd32(g[r],n-nmin[r]);
#endif
    }

    for (r = 0; r < subseq_Q; r++)
      if (nmin[r] != UINT32_MAX)
      {
        tmp[s].seq = i;
        tmp[s].d = r;
        tmp[s].M = make_bitmap(m_high-m_low+1);
#ifndef NDEBUG
        tmp[s].mcount = 0;
#endif
#if CHECK_FOR_GFN
        tmp[s].a = MAX(1,g[r]);
        tmp[s].b = nmin[r]%tmp[s].a;
#endif
        subseq[r] = s++;
      }

    assert (s <= s_count);

    SEQ[i].last = s-1;

    for (j = 0; j < SEQ[i].ncount; j++)
    {
      n = SEQ[i].N[j];
      r = n % subseq_Q;
      set_bit(tmp[subseq[r]].M, n/subseq_Q-m_low);
#ifndef NDEBUG
      tmp[subseq[r]].mcount++;
#endif
    }

    free(SEQ[i].N);
#ifndef NDEBUG
    old_count += SEQ[i].ncount;
#endif
  }

  assert (s == s_count);

#if (BASE == 0)
  report(1,"Split %"PRIu32" base %"PRIu32" sequence%s into %"PRIu32" base %"
         PRIu32"^%"PRIu32" subsequence%s.",seq_count,b_term,plural(seq_count),
         s_count,b_term,subseq_Q,plural(s_count));
#else
  report(1,"Split %"PRIu32" base %u sequence%s into %"PRIu32" base %u^%"PRIu32
         " subsequence%s.", seq_count, (unsigned int)BASE, plural(seq_count),
         s_count, (unsigned int)BASE, subseq_Q, plural(s_count));
#endif

  SUBSEQ = tmp;
  subseq_count = s_count;

#ifndef NDEBUG
  if (verbose_opt)
    report(1,"Used %"PRIu32" Kb for subsequence bitmaps.",
           subseq_count*(m_high-m_low)/8/1024);

  for (i = 0, t_count = 0; i < subseq_count; i++)
    t_count += SUBSEQ[i].mcount;
  assert(t_count == old_count);
#endif

  make_subseq_congruence_tables();
}

/* 0 if all terms have even n;
   1 if all terms have odd n;
   2 if some terms are even and some are odd.
*/
uint32_t seq_parity(uint32_t seq)
{
  uint32_t i, parity;

  assert(subseq_Q % 2 == 0);

  for (i = SEQ[seq].first, parity = SUBSEQ[i].d % 2; i <= SEQ[seq].last; i++)
    if (SUBSEQ[i].d % 2 != parity)
    {
#if ALLOW_MIXED_PARITY_SEQUENCES
      return 2;
#else
      error("Sequence %s has mixed parity terms.",seq_str(i));
#endif
    }

  return parity;
}

static void report_factor(uint32_t subseq, uint32_t n, uint64_t p)
{
  uint32_t k = SEQ[SUBSEQ[subseq].seq].k;
  int32_t c = SEQ[SUBSEQ[subseq].seq].c;

#if CHECK_FACTORS
  if (!is_factor(k,c,n,p))
    error("%"PRIu64" DOES NOT DIVIDE %s.",p,kbnc_str(k,n,c));
#endif

  factor_count++;
#ifndef NDEBUG
  SUBSEQ[subseq].mcount--;
#endif
  save_factor(k,c,n,p);
  report(1,"%"PRIu64" | %s",p,kbnc_str(k,n,c));
  notify_event(factor_found);
}

static void report_duplicate_factor(uint32_t seq, uint32_t n, uint64_t p)
{
  uint32_t k = SEQ[seq].k;
  int32_t c = SEQ[seq].c;

#if CHECK_FACTORS
  if (!is_factor(k,c,n,p))
    error("%"PRIu64" DOES NOT DIVIDE %s (duplicate).",p,kbnc_str(k,n,c));
#endif

  if (n <= n_max)
  {
    report(1,"%"PRIu64" | %s (duplicate)",p,kbnc_str(k,n,c));
    notify_event(factor_found);
  }
}

int benchmarking = 0;

/* eliminate a single term n=Qm+d for this subsequence.
*/
void eliminate_term(uint32_t subseq, uint32_t m, uint64_t p)
{
  assert(subseq < subseq_count);

  if (!benchmarking)
  {
    if (subseq_clear_m(subseq,m))
     report_factor(subseq,m*subseq_Q+SUBSEQ[subseq].d,p);
    else if (duplicates_opt)
     report_duplicate_factor(SUBSEQ[subseq].seq,m*subseq_Q+SUBSEQ[subseq].d,p);
  }
}
